#ifndef ASYM_GENERATIVE_HISTORY_H
#define ASYM_GENERATIVE_HISTORY_H
#include "Box.h"
namespace Asym{

	struct DecompCandidate{
		std::vector<int> indicesOfPartA;
		std::vector<int> indicesOfPartB;
		int dim;
		int pos;
	};

	typedef std::pair<int,int> OperationResult;
	typedef std::pair<Box,Box> PairBoxes;

	enum OperationType{OT_SplitX=0,OT_SplitY=1,OT_Delayer=2,OT_Undefined};

	struct HierarchyNode{
		Box box;
		int leftBoxId;
		int rightBoxId;
		int parentBoxId;
		std::vector<int> m_boxIndices;
		BoxSet contentBoxes;
		int sourcewidth;
		int sourceheight;
		HierarchyNode():leftBoxId(-1),rightBoxId(-1),parentBoxId(-1),otype(OT_Undefined){
		}

		HierarchyNode(const HierarchyNode& you){
			box=you.box;
			leftBoxId=you.leftBoxId;
			rightBoxId=you.rightBoxId;
			parentBoxId=you.parentBoxId;
			contentBoxes=you.contentBoxes;
			abstractLevel=you.abstractLevel;
			layer=you.layer;
			if (you.contentBoxes.size()>=1)
			{
				sourcewidth=you.contentBoxes[0].width();
				sourceheight=you.contentBoxes[0].height();
			}
			
			m_boxIndices=you.m_boxIndices;
		}
		OperationType otype;
		int para;
		int abstractLevel;
		int layer;
		void retargetContentData(int dx,int dy,double xscale,double yscale);
		
	};

	//GenHistķAnalysisʵˣôຯʵֻ
	//1.visualizationעԭʼķһöֱVisualizeĻ̫ˣ
	//   ĶһЩкϲ˵ڵһlevelǿհboxǽboxxзΣ
	//   ڶУͻĲˣӾϣǿԽǺϲһ𣬽һlevelͱbox
	
	// 2. ʵGenHist֮ƥ䣬ﲻǻԭʼƥ䣬ǻںϲ֮ƥ䣬
	//     ÿGenHistԿкܶlevelÿlevelʵһbox(һstep ǿհ׵boxȻɸbox...)
	//     ƥĹ̾GenHist AGenHist B֮levelƥ䣬ÿlevelƥboxsetƥ
	class GenerativeHistory{
		int _rootBoxId;
		std::vector<HierarchyNode*> _hierarchyBoxes;

		int computeLayersForSubTree(int rootNodeId,int startLayerId);
		HierarchyNode* allocateNewHierarchyNode();
		int updateHistoryWithBoundingBox_recursive(Box& box,int rid);
	public:
		GenerativeHistory();
		GenerativeHistory(const GenerativeHistory&);
		~GenerativeHistory();

		void saveHistory(char const * fn);
		void openHistory(char const *  fn);

		void updateHistoryWithBoundingBox(Box& box);

		int initialize(Box&);
		int rootBox() const;
		void setrootid(int id){_rootBoxId=id;}
		int leftBoxId(int boxId) const;
		int rightBoxId(int boxId) const;
		int parentBoxId(int boxId) const;
		int brotherNodeId(int nodeId) const;
		bool isRightChildOf(int nodeId,int pid) const;
		Box boxOf(int nodeId) const;
		int level(int nodeId) const;
		int levelNum() const;
		int leaveNodeNum() const; 
		bool isLeafNode(int nodeId) const;

		int layer(int nodeId) const;
		void setLayer(int nodeId,int L);
		int abstractLevel(int nodeId) const;
		void setAbstractLevel(int nodeId,int L);

		//Ϊԭʼķһö˵AѳA1A2boxesȻA2ַֽΪA21,A22boxes
		//ôvisualizeʱնôƣͷǳˣһЩϲ
		//ζxѣҾͰǷͬһlevel
		void computeAbstractionLevels();
		void computeLayers();
		int abstractionLevelNum();

		void boxesOfAbstractionLevels(std::vector<std::vector<int>>& bset);
		void boxesOfLayers(std::vector<std::vector<int>>& bset);
		void saveSvg(std::string );
		void saveSvg_AbstractionLevel(std::string);
		void debug();

		int layerNum();
		//void setNodeData(int nodeId,std::vector<int>& boxIndices);
		BoxSet& nodeData(int nodeId);
		
		void clear();
		int nodeNum() const;
		//HierarchyNode& nodeAt(int nodeId);

		GenerativeHistory& operator=(const GenerativeHistory&you);

		OperationType operationType(int nodeId);
		OperationResult split(int nodeId,int pos, int dim);
		OperationResult delayer(int nodeId,Box& newBoxA,Box& newBoxB);
		PairBoxes historyNode(int nodeId);
		

	//retargeting
		void retargeting(int nodeId,int dx,int dy,int dw,int dh);
		void retargetingInside(int rightChildId,int dx,int dy);
		//////////////////////////////////////////////////////////////////////////
		//for ga
		float nofnonleafnode();
		std::vector<int>getChildrenNode(int nodeId);
		void newhist(GenerativeHistory hist1,std::vector<int> selectlist);
		std::vector<int>AddirstNode(GenerativeHistory hist,std::vector<int> selectlist);
		void SetContentBoxesIndex(int index,int n,std::vector<int>list);
		void SetContentBoxesIndex(int index,std::vector<int>list);
		std::vector<int> GetContentBoxesIndex(int index);
		void AddSubtree(GenerativeHistory hist1,std::vector<int>nodelist,int parent);
		bool iseaqual(GenerativeHistory hist1,int hn0,int hn1);
		bool iseaqual(std::vector<int>list0,std::vector<int>list1);

		void setp(DecompCandidate dc,int index);

		//NIS
		/*float NIS(int index);
		float evaluateNIS(DecompCandidate& dc,Box space);*/
	

		//////////////////////////////////////////////////////////////////////////


		//==============================
		//кGenHistƥ

		static double nodeDistance(PairBoxes& nodeA,PairBoxes& nodeB);
		double findCorrespondenceToYou(GenerativeHistory& hist,std::vector<int>& mapToYou,std::vector<double> &distToYou);
		int findCorrespondenceToMe(PairBoxes& yourNode,double *minD=NULL);
		
		double distanceTo(GenerativeHistory& hist,int *bestLevelOfYou=NULL);
		double distanceTo(GenerativeHistory& hist,int myLevel,int yourLevel);
		//add
		double distanceTo2(GenerativeHistory& hist);
		double maxsimilarity(GenerativeHistory& hist0,GenerativeHistory& hist1);
		double anchoresimilarity(GenerativeHistory& hist0,int index0,GenerativeHistory& hist1,int index1,double diagonal);
		double dist(HierarchyNode& hn0,HierarchyNode& hn1,double diagonal);
		void normalize();
		void normalize0();
		void normalize2();
		Box getboundingbox(HierarchyNode& hn0);
		
		void  nodeAtLevel(std::vector<std::vector<int>>& tmp);


		void normalizedHistory(int x,int y, float scaleRatio);
		void scaleHistory(int rootNodeId,Box& newRootBoxSize);
		void translateHistory(int rootNodeId,int dx,int dy);

		static void historyNode_saveSvg(PairBoxes& nodeA,std::string fn);

		Box evaluateSizeForTreeVisualization();
		std::vector<std::vector<int>>  computeXPosForLevel(int startX,std::vector<std::vector<int>> &levelNodes);

		void saveSvg_Tree_root(std::string fn,bool showContent=true);
		void saveSvg_Tree(std::string fn,bool showContent=true);
		void saveSvg_TreeforGCut(std::string fn,std::vector<int>edge,bool showContent=true);
		void saveSvg_TreewithNIS(std::string fn,std::vector<float>nis,std::vector<int>nlist,bool showContent=true);
		std::vector<int>noleaflist()
		{
			std::vector<int>nlist;
			nlist.clear();
			for (int i=0;i<nodeNum();i++)
			{
				if (!isLeafNode(i))
				{
					nlist.push_back(i);
				}
			}
			return nlist;
		}


#define TreeMatchFlag_A2B 1
#define TreeMatchFlag_B2A 2
		static void saveSvg_TreeMatch(std::string fn,GenerativeHistory& histA,GenerativeHistory& histB,std::vector<int> &correspondA2B,std::vector<int> &correspondB2A,int flag);


		int findContainerNode(int x,int y);
		int findContainerNode(Box);
		void contentBoxesFromLeaveNodes(BoxSet &bset);
	};


}
#endif